@@ -12,6 +12,7 @@ use Math::BigInt;
 use IO::Socket::INET;
 use IO::Socket::UNIX;
 use Encode qw/encode_utf8 decode_utf8/;
+use List::MoreUtils qw/any/;
 
 my $is_native64 = $Config{longsize} == 8 || defined $Config{use64bitint} || defined $Config{use64bitall};
     
@@ -24,7 +25,11 @@ Sphinx::Search - Sphinx search engine API Perl client
 
 Please note that you *MUST* install a version which is compatible with your version of Sphinx.
 
-Use version 0.28 for Sphinx-2.0.3-release (svn-r3043)
+Use version 0.29 for Sphinx-2.2.8-release or later (or use DBI instead)
+
+Use version 0.28 for Sphinx-2.0.8-release or later
+
+Use version 0.27.2 for Sphinx-2.0.3-release (svn-r3043)
 
 Use version 0.26.1 for Sphinx-2.0.1-beta (svn-r2792)
 
@@ -56,7 +61,7 @@ Use version 0.02 for Sphinx 0.9.8-cvs-20070818
 
 =cut
 
-our $VERSION = '0.28';
+our $VERSION = '0.29';
 
 =head1 SYNOPSIS
 
@@ -64,15 +69,34 @@ our $VERSION = '0.28';
 
     $sph = Sphinx::Search->new();
 
-    $results = $sph->SetMatchMode(SPH_MATCH_ALL)
-                   ->SetSortMode(SPH_SORT_RELEVANCE)
+    # Standard API query
+    $results = $sph->SetSortMode(SPH_SORT_RELEVANCE)
                    ->Query("search terms");
 
+    # SphinxQL query
+    $results = $sph->SphinxQL("SELECT * FROM myindex WHERE MATCH('search terms')");
+
 =head1 DESCRIPTION
 
 This is the Perl API client for the Sphinx open-source SQL full-text indexing
 search engine, L<http://www.sphinxsearch.com>.
 
+Since 0.9.9, Sphinx supports a native MySQL-protocol client, i.e. DBI with DBD::mysql. That is, you can configure the server to have a mysql41 listening port and then simply do
+
+  my $dbh = DBI->connect('dbi:mysql:host=127.0.0.1;port=9306;mysql_enable_utf8=1') or die "Failed to connect via DBI";
+  my $sth = $dbh->prepare_cached("SELECT * FROM myindex WHERE MATCH('search terms')");
+  $sth->execute();
+  while (my $row = $sth->fetchrow_arrayref) {
+      ... # Collect results
+  }
+
+The DBI client turns out to be significantly (about 5x) faster than this pure-Perl API. You should probably be using that instead.
+
+This module also supports SphinxQL queries, with the small advantage that you can use either the standard API or the SphinxQL API over the one port (i.e. the searchd server does not need to be configured with a mysql41 listening port).
+
+Given that the DBI client has several advantages over this API, future updates of this module are unlikely.
+
+  
 =cut
 
 # Constants to export.
@@ -88,6 +112,8 @@ our @EXPORT = qw(
                 SPH_GROUPBY_ATTRPAIR
                 SPH_ATTR_INTEGER SPH_ATTR_TIMESTAMP SPH_ATTR_ORDINAL SPH_ATTR_BOOL
                 SPH_ATTR_FLOAT SPH_ATTR_BIGINT SPH_ATTR_STRING SPH_ATTR_MULTI SPH_ATTR_MULTI64
+                SPH_QF_REVERSE_SCAN SPH_QF_SORT_METHOD SPH_QF_MAX_PREDICTED_TIME 
+                SPH_QF_BOOLEAN_SIMPLIFY SPH_QF_IDF SPH_QF_GLOBAL_IDF
                 );
 
 # known searchd commands
@@ -97,15 +123,17 @@ use constant SEARCHD_COMMAND_UPDATE     => 2;
 use constant SEARCHD_COMMAND_KEYWORDS   => 3;
 use constant SEARCHD_COMMAND_PERSIST    => 4;
 use constant SEARCHD_COMMAND_STATUS     => 5;
-use constant SEARCHD_COMMAND_FLUSHATTRS         => 7;
+use constant SEARCHD_COMMAND_FLUSHATTRS => 7;
+use constant SEARCHD_COMMAND_SPHINXQL   => 8;
 
 # current client-side command implementation versions
-use constant VER_COMMAND_SEARCH                 => 0x119;
+use constant VER_COMMAND_SEARCH                 => 0x11E;
 use constant VER_COMMAND_EXCERPT        => 0x104;
-use constant VER_COMMAND_UPDATE                 => 0x102;
+use constant VER_COMMAND_UPDATE                 => 0x103;
 use constant VER_COMMAND_KEYWORDS       => 0x100;
-use constant VER_COMMAND_STATUS         => 0x100;
+use constant VER_COMMAND_STATUS         => 0x101;
 use constant VER_COMMAND_FLUSHATTRS     => 0x100;
+use constant VER_COMMAND_SPHINXQL       => 0x100;
 
 # known searchd status codes
 use constant SEARCHD_OK                         => 0;
@@ -146,6 +174,7 @@ use constant SPH_SORT_EXPR              => 5;
 use constant SPH_FILTER_VALUES          => 0;
 use constant SPH_FILTER_RANGE           => 1;
 use constant SPH_FILTER_FLOATRANGE      => 2;
+use constant SPH_FILTER_STRING          => 3;
 
 # known attribute types
 use constant SPH_ATTR_INTEGER           => 1;
@@ -155,6 +184,7 @@ use constant SPH_ATTR_BOOL              => 4;
 use constant SPH_ATTR_FLOAT             => 5;
 use constant SPH_ATTR_BIGINT            => 6;
 use constant SPH_ATTR_STRING            => 7;
+use constant SPH_ATTR_FACTORS           => 1001;
 use constant SPH_ATTR_MULTI             => 0x40000001;
 use constant SPH_ATTR_MULTI64           => 0x40000002;
 
@@ -166,6 +196,25 @@ use constant SPH_GROUPBY_YEAR           => 3;
 use constant SPH_GROUPBY_ATTR           => 4;
 use constant SPH_GROUPBY_ATTRPAIR       => 5;
 
+use constant {
+    SPH_QF_REVERSE_SCAN => 'reverse_scan',
+    SPH_QF_SORT_METHOD => 'sort_method',
+    SPH_QF_MAX_PREDICTED_TIME => 'max_predicted_time',
+    SPH_QF_BOOLEAN_SIMPLIFY => 'boolean_simplify',
+    SPH_QF_IDF => 'idf',
+    SPH_QF_GLOBAL_IDF => 'global_idf',
+};
+
+my %query_flags = ( SPH_QF_REVERSE_SCAN() => [ 0, 1 ],
+                    SPH_QF_SORT_METHOD() => [ qw/pq kbuffer/ ],
+                    SPH_QF_MAX_PREDICTED_TIME() => [ 0 ],
+                    SPH_QF_BOOLEAN_SIMPLIFY() => [1, 0],
+                    SPH_QF_IDF() => [ qw/normalized plain tfidf_normalized tfidf_unnormalized/ ],
+                    SPH_QF_GLOBAL_IDF() => [1, 0],
+                );
+
+use constant MYSQL_COL_STRING => 254;
+
 # Floating point number matching expression
 my $num_re = qr/^-?\d*\.?\d*(?:[eE][+-]?\d+)?$/;
 
@@ -244,6 +293,17 @@ sub _sphUnpackI64
     return $x->bstr;
 }
 
+sub _sphSetBit {
+    my ($self, $flag, $bit, $on) = @_;
+
+    if ($on) {
+        $flag |= (1 << $bit);
+    }
+    else {
+        $flag &= ~(1 << $bit);
+    }
+    return $flag;
+}
 
 
 
@@ -293,7 +353,7 @@ sub new {
         # per-query settings
         _offset                 => 0,
         _limit          => 20,
-        _mode           => SPH_MATCH_ALL,
+        _mode           => SPH_MATCH_EXTENDED2,
         _weights        => [],
         _sort           => SPH_SORT_RELEVANCE,
         _sortby                 => "",
@@ -331,6 +391,9 @@ sub new {
     };
     bless $self, ref($class) || $class;
 
+    $self->ResetQueryFlag;
+    $self->ResetOuterSelect;
+
     # These options are supported in the constructor, but not recommended 
     # since there is no validation.  Use the Set* methods instead.
     my %legal_opts = map { $_ => 1 } qw/host port offset limit mode weights sort sortby groupby groupbyfunc maxmatches cutoff retrycount retrydelay log debug string_encoder string_decoder/;
@@ -356,6 +419,12 @@ sub _Error {
     return;
 }
 
+sub _Throw {
+    my ($self, $msg) = @_;
+
+    die $msg;
+}
+
 =head2 GetLastError
 
     $error = $sph->GetLastError;
@@ -480,10 +549,13 @@ sub SetServer {
         return;
     }
         
-    croak("port is not an integer") unless defined($port) && $port =~ m/^\d+$/o;
+    $port ||= 0;
+    croak("port is not an number") unless $port =~ m/^\d+/o;
+    $port = int($port);
+    croak("port $port out of range 0 to 65536") if $port <0 || $port >= 65536;
 
     $self->{_host} = $host;
-    $self->{_port} = $port;
+    $self->{_port} = $port == 0 ? 9312 : $port;
     $self->{_path} = undef;
 
     return $self;
@@ -773,6 +845,8 @@ sub SetMaxQueryTime {
 
 =head2 SetMatchMode
 
+** DEPRECATED **
+
     $sph->SetMatchMode($mode);
 
 Set match mode, which may be one of:
@@ -809,6 +883,8 @@ Returns $sph.
 sub SetMatchMode {
     my $self = shift;
     my $mode = shift;
+    warn "SetMatchMode is DEPRECATED. Do not call this method - use extended query syntax instead.";
+
     croak("Match mode not defined") unless defined($mode);
     croak("Unknown matchmode: $mode") unless ( $mode==SPH_MATCH_ALL 
                                                || $mode==SPH_MATCH_ANY 
@@ -874,8 +950,9 @@ sub SetRankingMode {
     my $ranker = shift;
     my $rankexp = shift;
 
-    croak("Unknown ranking mode: $ranker") unless ( $ranker >= 0
-                                                    && $ranker < SPH_RANK_TOTAL );
+    croak("Unknown ranking mode: $ranker") unless ( $ranker == 0
+						    || ( $ranker >= 1
+							 && $ranker < SPH_RANK_TOTAL ));
 
     $self->{_ranker} = $ranker;
     $self->{_rankexpr} = $rankexp || "";
@@ -937,6 +1014,8 @@ sub SetSortMode {
 
 =head2 SetWeights
     
+** DEPRECATED **
+
     $sph->SetWeights([ 1, 2, 3, 4]);
 
 This method is deprecated.  Use L<SetFieldWeights> instead.
@@ -951,6 +1030,8 @@ Returns $sph.
 sub SetWeights {
     my $self = shift;
     my $weights = shift;
+    warn "SetWeights is DEPRECATED, Do not call this method; use SetFieldWeights instead";
+
     croak("Weights is not an array reference") unless (ref($weights) eq 'ARRAY');
     foreach my $weight (@$weights) {
         croak("Weight: $weight is not an integer") unless ($weight =~ /^\d+$/);
@@ -1037,7 +1118,7 @@ sub SetIDRange {
     $sph->SetFilter($attr, \@values, $exclude);
 
 Sets the results to be filtered on the given attribute.  Only results which have
-attributes matching the given values will be returned.
+attributes matching the given values will be returned.  (Attribute values must be integers).
 
 This may be called multiple times with different attributes to select on
 multiple attributes.
@@ -1065,6 +1146,31 @@ sub SetFilter {
     return $self;
 }
 
+=head2 SetFilterString
+
+    $sph->SetFilterString($attr, $value)
+    $sph->SetFilterString($attr, $value, $exclude)
+
+Adds new string value filter.  Only those documents where $attr column value matches the string value from $value will be matched (or rejected, if $exclude is true).
+
+=cut
+
+sub SetFilterString {
+    my ($self, $attribute, $value, $exclude) = @_;
+
+    croak("attribute is not defined") unless (defined $attribute);
+    croak("value is not a string") unless ($value && ! ref($value));
+
+    push(@{$self->{_filters}}, {
+        type => SPH_FILTER_STRING,
+        attr => $attribute,
+        value => $value,
+        exclude => $exclude ? 1 : 0,
+         });
+
+    return $self;
+}
+
 =head2 SetFilterRange
 
     $sph->SetFilterRange($attr, $min, $max);
@@ -1307,6 +1413,8 @@ sub SetRetries {
 
 =head2 SetOverride
 
+ ** DEPRECATED **
+
     $sph->SetOverride($attrname, $attrtype, $values);
 
  Set attribute values override. There can be only one override per attribute.
@@ -1320,6 +1428,8 @@ sub SetOverride {
     my $attrtype = shift;
     my $values = shift;
 
+    die "SetOverride is DEPRECATED. Do not call this method.";
+
     croak("attribute name is not defined") unless defined $attrname;
     croak("Uknown attribute type: $attrtype") unless ($attrtype == SPH_ATTR_INTEGER
                                                       || $attrtype == SPH_ATTR_TIMESTAMP
@@ -1349,6 +1459,69 @@ sub SetSelect {
     return $self;
 }
 
+=head2 SetQueryFlag 
+
+    $sph->SetQueryFlag($flag_name, $flag_value);
+
+
+=cut
+
+sub SetQueryFlag {
+    my ($self, $flag_name, $flag_value) = @_;
+
+    croak("Unknown flag $flag_name") unless exists $query_flags{$flag_name};
+    croak("Unknown or illegal flag value ($flag_value) for '$flag_name'") unless 
+        (any { $_ eq $flag_value } @{$query_flags{$flag_name}})
+            || ($flag_name eq 'max_predicted_time' && $flag_value =~ m/^\d+$/);
+
+    if ($flag_name eq SPH_QF_REVERSE_SCAN) {
+        $self->{_query_flags} = $self->_sphSetBit( $self->{_query_flags}, 0, $flag_value == 1);
+    }
+    elsif ($flag_name eq SPH_QF_SORT_METHOD) {
+        $self->{_query_flags} = $self->_sphSetBit( $self->{_query_flags}, 1, $flag_value == "kbuffer");
+    }
+    elsif ($flag_name eq SPH_QF_MAX_PREDICTED_TIME) {
+        $self->{_query_flags} = $self->_sphSetBit( $self->{_query_flags}, 2, $flag_value > 0);
+        $self->{_predictedtime} = $flag_value;
+    }
+    elsif ($flag_name eq SPH_QF_BOOLEAN_SIMPLIFY) {
+        $self->{_query_flags} = $self->_sphSetBit( $self->{_query_flags}, 3, $flag_value);
+    }
+    elsif ($flag_name eq SPH_QF_IDF) {
+        if ($flag_value eq 'normalized' || $flag_value eq 'plain') {
+            $self->{_query_flags} = $self->_sphSetBit( $self->{_query_flags}, 4, $flag_value eq 'normalized');
+        }
+        else { # must be tfidf_normalized or tfidf_unnormalized
+            $self->{_query_flags} = $self->_sphSetBit( $self->{_query_flags}, 6, $flag_value eq 'tfidf_normalized');
+        }
+    }
+    elsif ($flag_name eq SPH_QF_GLOBAL_IDF) {
+        $self->{_query_flags} = $self->_sphSetBit( $self->{_query_flags}, 5, $flag_value);
+    }
+
+    return $self;
+}
+
+=head2 SetOuterSelect
+
+  $sph->SetOuterSelect($orderby, $offset, $limit)
+
+=cut
+
+sub SetOuterSelect {
+    my ($self, $orderby, $offset, $limit) = @_;
+
+    croak("orderby must be a string") unless $orderby && ! ref($orderby);
+    croak("offset and limit must be integers > 0") unless $offset =~ m/^\d+$/ && $limit =~ m/^\d+$/;
+
+    $self->{_outerorderby} = $orderby;
+    $self->{_outeroffsetlimit} = $offset;
+    $self->{_outerlimit} = $limit;
+    $self->{_hasouter} = 1;
+
+    return $self;
+}
+
 =head2 ResetFilters
 
     $sph->ResetFilters;
@@ -1398,6 +1571,38 @@ sub ResetOverrides {
     return $self;
 }
 
+=head2 ResetQueryFlag
+
+Clear all query flags.
+
+=cut
+
+sub ResetQueryFlag {
+    my $self = shift;
+
+    $self->{_query_flags} = $self->_sphSetBit(0, 6, 1);
+    $self->{_predictedtime} = 0;
+    return $self;
+}
+
+=head2 ResetOuterSelect
+
+Clear all outer select settings.
+
+=cut
+
+sub ResetOuterSelect {
+    my $self = shift;
+
+    $self->{_outerorderby} = '';
+    $self->{_outeroffset} = 0;
+    $self->{_outerlimit} = 0;
+    $self->{_hasouter} = 0;
+
+    return $self;
+}
+
+
 =head2 Query
 
     $results = $sph->Query($query, $index);
@@ -1501,7 +1706,7 @@ sub AddQuery {
     ##################
 
     my $req;
-    $req = pack ( "NNNN", $self->{_offset}, $self->{_limit}, $self->{_mode}, $self->{_ranker}); # mode and limits
+    $req = pack ( "NNNNN", $self->{_query_flags}, $self->{_offset}, $self->{_limit}, $self->{_mode}, $self->{_ranker}); # mode and limits
 
     if ($self->{_ranker} == SPH_RANK_EXPR) {
         $req .= pack ( "N/a*", $self->{_rankexpr});
@@ -1531,6 +1736,9 @@ sub AddQuery {
         elsif ($t == SPH_FILTER_FLOATRANGE) {
             $req .= _PackFloat ( $filter->{"min"} ) . _PackFloat ( $filter->{"max"} );
         }
+        elsif ($t == SPH_FILTER_STRING) {
+            $req .= pack ( "N/a*", $filter->{value});
+        }
         else {
             croak("Unhandled filter type $t");
         }
@@ -1593,6 +1801,15 @@ sub AddQuery {
     # select list
     $req .= pack("N/a*", $self->{_select} || '');
 
+    # max_predicted_time
+    if ($self->{_predictedtime} > 0) {
+        $req .= pack ( "N", $self->{_predictedtime} );
+    }
+
+    $req .= pack ( "N/a*", $self->{_outerorderby});
+    $req .= pack ( "NN", $self->{_outeroffset}, $self->{_outerlimit} );
+    $req .= pack ("N", $self->{_hasouter} ? 1 : 0 );
+
     push(@{$self->{_reqs}}, $req);
 
     return scalar $#{$self->{_reqs}};
@@ -1727,11 +1944,11 @@ sub RunQueries {
                         push(@{$data->{$attr}}, $val);
                     }
                 }
-        elsif ($attrs{$attr} == SPH_ATTR_MULTI64) {
+                elsif ($attrs{$attr} == SPH_ATTR_MULTI64) {
                     my $nvalues = $val;
                     $data->{$attr} = [];
                     while ($nvalues > 0 && $p < $max) {
-                        $val = unpack( "Q*", substr ( $response, $p, 8 ) ); $p += 8;
+                        $val = unpack( "q*", substr ( $response, $p, 8 ) ); $p += 8;
                         push(@{$data->{$attr}}, $val);
                         $nvalues -= 2;
                     }
@@ -1740,6 +1957,10 @@ sub RunQueries {
                     $data->{$attr} = $self->{_string_decoder}->(substr ($response, $p, $val));
                     $p += $val;
                 }
+                elsif ($attrs{$attr} == SPH_ATTR_FACTORS) {
+                    $data->{$attr} = $self->{_string_decoder}->(substr ($response, $p, $val - 4));
+                    $p += $val - 4;
+                }
                 else {
                     $data->{$attr} = $val;
                 }
@@ -1768,6 +1989,206 @@ sub RunQueries {
     return \@results;
 }
 
+=head2 SphinxQL
+
+  my $results = $sph->SphinxQL($sphinxql_query);
+
+This is an alternative implementation of the SphinxQL API to the DBI option. Frankly, it was an experiment, and the DBI driver proved to have much better performance. Whilst this may be useful to some, in general if you are considering using this method then you should probably look at connecting directly via DBI instead.
+
+Results are return in a hash containing an array of 'columns' and 'rows' and possibly a warning count. If a server-side error occurs, the hash contains the 'error' field. If a communication error occurs, the return value will be undefined. In either error case, GetLastError will return the error.
+
+=cut
+
+sub SphinxQL {
+    my ($self, $ql) = @_;
+
+    my $req = pack ("N/a*", $self->{_string_encoder}->($ql));
+    $req = pack ( "nnN/a*", SEARCHD_COMMAND_SPHINXQL, VER_COMMAND_SPHINXQL, $req); # add header
+    my $response = $self->_ProcessRequest($req, VER_COMMAND_SPHINXQL);
+
+    return if ! $response;
+
+    my ($result) = $self->_mysql_unpack_table($response, 0, length($response));
+
+    return $result;
+}
+
+sub _mysql_unpack_header {
+    my ($self, $response, $p) = @_;
+
+    my $max = length($response);
+    return (undef, undef, $p, 1, undef, undef) if $p == $max;
+    $self->_Throw("Decode pointer ($p) beyond end of buffer ($max)") if $p > $max;
+
+    my $header = unpack ( "L<", substr( $response, $p, 4 ) ); $p += 4; # packet ID << 24 | packet length
+    my $packet_id = $header >> 24;
+    my $packet_len = $header & 0x00FFFFFF;
+    return ($packet_id, $packet_len, $p, 1, undef, undef) if $packet_len == 0;
+
+    # Peek for EOF or error
+    my ($is_eof, $warns, $more_results);
+    my $peek_byte = unpack("C", substr($response, $p, 1));
+    if ($packet_len == 5 && $peek_byte == 0xfe) { # EOF
+        $is_eof = 1;
+        $p += 1;                # 0xFE
+        my $warns = unpack ( "L<", substr( $response, $p, 4 ) ); $p += 4;
+        my $more_results = (($warns >> 16) & 8) != 0; # SPH_MYSQL_FLAG_MORE_RESULTS = 8
+        $warns &= 0xFFFF;
+    }
+    elsif ($peek_byte == 0xff) {
+        $p += 1;                # 0xFF
+        my $error_code = unpack ( "S<", substr( $response, $p, 2 ) ); $p += 2;
+        my $error_code2 = unpack("a*", substr($response, $p, 6)); $p += 6;
+        my $len = $packet_len - 9;
+        my $err_msg = $self->{_string_decoder}->(unpack("a*", substr($response, $p, $len))); $p += $len;
+        $self->_Throw("Error Code $error_code: $error_code2 $err_msg");
+    }
+    return ($packet_id, $packet_len, $p, $is_eof, $warns, $more_results);
+}
+
+sub _mysql_unpack_table {
+    my ($self, $response, $p, $max) = @_;
+
+    my ($columns, $warns, $more_results, $rows);
+
+    eval {
+        ($columns, $warns, $more_results, $p) = $self->_mysql_unpack_table_header($response, $p);
+        ($rows, $p, $warns, $more_results) = $self->_mysql_unpack_table_rows($response, $p, $max, $columns);
+    };
+    if (my $e = $@) {
+        $self->_Error($e);
+        return ({ error => $e }, $p);
+    }
+
+    my %result = (columns => $columns,
+                  rows => $rows,
+                  warnings => $warns,
+              );
+    return (\%result, $p);
+}
+
+sub _mysql_unpack_table_header {
+    my ($self, $response, $p) = @_;
+
+    my @columns;
+
+    # table header begin
+    my ($packet_id, $packet_len, $is_eof, $warns, $more_results);
+    ($packet_id, $packet_len, $p, $is_eof, $warns, $more_results) = $self->_mysql_unpack_header($response, $p);
+
+    if (! $is_eof) {
+        my $ncols;
+        ($ncols, $p) = $self->_mysql_unpack_varint($response, $p);
+
+        # column info
+        for my $i (0 .. $ncols - 1) {
+            ($columns[$i], $p) = $self->_mysql_unpack_field_packet($response, $p);
+        }
+
+        # table header end
+        ($packet_id, $packet_len, $p, $is_eof, $warns, $more_results) = $self->_mysql_unpack_header($response, $p);
+    }
+    return (\@columns, $warns, $more_results, $p);
+}
+
+sub _mysql_unpack_table_rows {
+    my ($self, $response, $p, $max, $columns) = @_;
+
+    my @rows;
+    my ($warns, $more_results);
+    while ($p < $max) {
+        my $row;
+        ($row, $p, $warns, $more_results) = $self->_mysql_unpack_table_row($response, $p, $columns);
+        if ($row) {
+            push(@rows, $row);
+        }
+        else {
+            last;
+        }
+    }
+    return (\@rows, $p, $warns, $more_results);
+}
+
+sub _mysql_unpack_table_row {
+    my ($self, $response, $p, $columns) = @_;
+
+    my ($packet_id, $packet_len, $is_eof, $warns, $more_results);
+    ($packet_id, $packet_len, $p, $is_eof, $warns, $more_results) = $self->_mysql_unpack_header($response, $p);
+    if ($is_eof) {
+        return(undef, $p, $warns, $more_results);
+    }
+
+    my @row;
+    for my $col (@$columns) {
+        my $val;
+        if ($col->{column_type} == MYSQL_COL_STRING) {
+            ($val, $p) = $self->_mysql_unpack_string($response, $p, 1);
+        }
+        else {
+            ($val, $p) = $self->_mysql_unpack_string($response, $p);
+        }
+        push(@row, $val);
+    }
+    return (\@row, $p);
+}
+
+
+sub _mysql_unpack_varint {
+    my ($self, $response, $p) = @_;
+
+    my $prefix = unpack("C", substr($response, $p, 1)); $p += 1;
+    if ($prefix < 251) {
+        return ($prefix, $p);
+    }
+    if ($prefix == 0xFC) {
+        $prefix = unpack("S<", substr($response, $p, 2)); $p += 2;
+        return ($prefix, $p);
+    }
+    if ($prefix == 0xFC) {
+        $prefix = unpack("S<", substr($response, $p, 2)); $p += 2;
+        $prefix += unpack("C", substr($response, $p, 1)) << 16; $p += 1;
+        return ($prefix, $p);
+    }
+    if ($prefix == 0xFE) {
+        $prefix = unpack("L<", substr($response, $p, 4)); $p += 4;
+    }
+    $p += 4; # discard 4 null bytes
+    return ($prefix, $p);
+}
+
+sub _mysql_unpack_string {
+    my ($self, $response, $p, $decode) = @_;
+
+    my $len;
+    ($len, $p) = $self->_mysql_unpack_varint($response, $p);
+    my $s = substr($response, $p, $len); $p += $len;
+    $s = $self->{_string_decoder}->($s) if $decode;
+
+    return ($s, $p);
+}
+
+sub _mysql_unpack_field_packet {
+    my ($self, $response, $p) = @_;
+
+    my ($packet_id, $packet_len, $is_eof, $warns, $more_results);
+    ($packet_id, $packet_len, $p, $is_eof, $warns, $more_results) = $self->_mysql_unpack_header($response, $p);
+
+    my %field;
+    ($field{catalog}, $p) = $self->_mysql_unpack_string($response, $p);
+    ($field{db}, $p) = $self->_mysql_unpack_string($response, $p);
+    ($field{table}, $p) = $self->_mysql_unpack_string($response, $p);
+    ($field{org_table}, $p) = $self->_mysql_unpack_string($response, $p);
+    ($field{name}, $p) = $self->_mysql_unpack_string($response, $p);
+    ($field{org_name}, $p) = $self->_mysql_unpack_string($response, $p);
+
+    $p += 3; # filler=12, charset_nr=0x21 (utf8)
+    $field{column_length} = unpack ( "L<", substr( $response, $p, 4 ) ); $p += 4;
+    $field{column_type} = unpack("C", substr($response, $p, 1)); $p += 1;
+    $p += 5; # flags, decimals, filler
+
+    return (\%field, $p);
+}
+
 =head2 BuildExcerpts
 
     $excerpts = $sph->BuildExcerpts($docs, $index, $words, $opts)
@@ -2043,6 +2464,7 @@ sub EscapeString {
 
     $sph->UpdateAttributes($index, \@attrs, \%values);
     $sph->UpdateAttributes($index, \@attrs, \%values, $mva);
+    $sph->UpdateAttributes($index, \@attrs, \%values, $mva, $ignorenonexistent);
 
 Update specified attributes on specified documents
 
@@ -2060,6 +2482,14 @@ Array of attribute name strings
 
 A hash with key as document id, value as an array of new attribute values
 
+=item mva
+
+If set, indicates that there is update of MVA attributes
+
+=item ignorenonexistent
+
+If set, the update will silently ignore any warnings about trying to update a column which is not exists in current index schema.
+
 =back
 
 Returns number of actually updated documents (0 or more) on success
@@ -2073,7 +2503,7 @@ Usage example:
 =cut
 
 sub UpdateAttributes  {
-    my ($self, $index, $attrs, $values, $mva ) = @_;
+    my ($self, $index, $attrs, $values, $mva, $ignorenonexistent ) = @_;
 
     croak("index is not defined") unless (defined $index);
     croak("attrs must be an array") unless ref($attrs) eq "ARRAY";
@@ -2104,6 +2534,7 @@ sub UpdateAttributes  {
     my $req = pack ( "N/a*", $index);
 
     $req .= pack ( "N", scalar @$attrs );
+    $req .= pack ( "N", $ignorenonexistent ? 1 : 0 );
     for my $attr (@$attrs) {
         $req .= pack ( "N/a*", $attr)
             . pack("N", $mva ? 1 : 0);
@@ -2192,6 +2623,7 @@ sub Close {
 =head2 Status
 
     $status = $sph->Status()
+    $status = $sph->Status($session)
 
 Queries searchd status, and returns a hash of status variable name and value pairs. 
 
@@ -2201,9 +2633,9 @@ Returns undef on failure.
 
 sub Status {
     
-    my $self = shift;
+    my ($self, $session) = @_;
 
-    my $req = pack("nnNN", SEARCHD_COMMAND_STATUS, VER_COMMAND_STATUS, 4, 1 ); # len=4, body=1
+    my $req = pack("nnNN", SEARCHD_COMMAND_STATUS, VER_COMMAND_STATUS, 4, $session ? 0 : 1 ); # len=4, body=1
 
     my $response = $self->_ProcessRequest($req, VER_COMMAND_STATUS);
     return unless $response;
@@ -2329,7 +2761,7 @@ Thanks to Alexey Kholodkov for contributing a significant patch for handling per
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2012 Jon Schutz, all rights reserved.
+Copyright 2015 Jon Schutz, all rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License.
